% Inviscid-flow model predicting evolution of a train of vortical structures
% Developed by Tim Berk, 2018
% Contact: BerkTimA@gmail.com
% Originally used in Berk, T. and Ganapathisubramani, B. 2019 Effects of
% vortex-induced velocity on the development of a synthetic jet issuing
% into a turbulent boundary layer. J. Fluid Mech. 

%% initial conditions
St = 0.5;   % Strouhal number as defined in Berk and Ganapathisubramani (2019)
r = 0.88;   % velocity ratio as defined in Berk and Ganapathisubramani (2019)
Uinf = 10;  % freestream velocity (m/s), used in definition of St
l = 13E-3;  % slot width (m), used in definition of St
d = 0.0454; % BL thickness

a = 35; % inclination angle of filaments (deg)
D = 0.020; % diameter of filament rings

G = [0.04 -0.06]; % starting circulation of up- and downstream sides of the vortex ring (m^2/s)

adv_fact = 0.8; % advection factor (compared to Uinf)

x0 = d; % streamwise starting position of center of the ring
y0 = 0.6*d; % wall-normal starting position of center of the ring

% model parameters:
n = 20; % approximate nodes per filament head
R = 0.01; % Rankine vortex radius
dtT = 1/8;  % time-step
tT_max = 5; % number of cycles
tT = 1:dtT:tT_max;  % time-series

%% Generate filaments
[X_nod,Y_nod,Z_nod,G_seg] = Filament_Geometry(a,D,x0,y0,G,n); % generates filament, function embedded at bottom of the script
G_seg = [G_seg; ones(1,size(G_seg,2))]; % adds structure number

n_cycle = 1; % counter for number of cycles
f1 = figure;
for t = 1:length(tT)
    % create new structure each cycle
    if tT(t)-n_cycle >= 1 % at start of new cycle
        n_cycle = n_cycle + 1; % increase cycle counter
        [X_nod_new,Y_nod_new,Z_nod_new,G_seg_new] = Filament_Geometry(a,D,x0,y0,G,n); % generates filament, function embedded at bottom of the script
        X_nod = [X_nod X_nod_new]; % adds new information on filament position
        Y_nod = [Y_nod Y_nod_new]; % adds new information on filament position
        Z_nod = [Z_nod Z_nod_new]; % adds new information on filament position
        G_seg = [G_seg [G_seg_new; round(tT(t))*ones(1,size(G_seg_new,2))]]; % adds new information on filament circulation
    end
    % let circulation decay at each time step
    G_sign = sign(G_seg(1,:)); % save sign of circulation (positive or negative)
    G_temp = (abs(G_seg(1,:)) - 1.61 * dtT * d/(St*Uinf)); % linear decay of circulation with dG/dt = 1.61
    G_temp(G_temp<0) = 0;   % decaying circulation cannot change sign
    G_seg(1,:) = G_sign.*G_temp; % add sign
    for i = min(G_seg(3,:)):max(G_seg(3,:)) % loop through each structure
        % fix the leg to still be difference between upstream and downstream
        % part of the vortex ring, i.e. this defines decay of the leg:
        i_p = find(G_seg(2,:)==1 & G_seg(3,:)==i,1,'first');
        i_n = find(G_seg(2,:)==-1 & G_seg(3,:)==i,1,'first');
        i_l_0 = find(G_seg(2,:)==0 & G_seg(3,:)==i,1,'first');
        i_l_1 = find(G_seg(2,:)==0 & G_seg(3,:)==i,1,'last');
        G_l = G_seg(1,i_p)-G_seg(1,i_n);
        n_l = i_l_1-i_l_0+1;
        G_seg(1,i_l_0:i_l_1) = repmat(G_l,1,n_l);   % uniform circulation along leg
        G_seg(1,i_l_0:i_l_1) = linspace(G_l,0,n_l);   % decaying circulation in leg
        
        % delete filaments which have zero circulation
        if max(abs(G_seg(1,G_seg(3,:)==i)))==0
            X_nod(:,G_seg(3,:)==i) = [];
            Y_nod(:,G_seg(3,:)==i) = [];
            Z_nod(:,G_seg(3,:)==i) = [];
            G_seg(:,G_seg(3,:)==i) = [];
        end
    end
    
    %% describe data points - points at which induced velocity will be calculated
    % here this is only at locations of filaments
    x = X_nod;
    y = Y_nod;
    z = Z_nod;
    
    % the model uses reflection at the wall. the model also uses symmetry
    % around z=0
    x_refl = [x x x x];
    y_refl = [y y -y -y];
    z_refl = [z -z z -z];
    
    %% apply Biot-Savart
    [Ux_refl,Uy_refl,Uz_refl] = Biot_Savart(X_nod,Y_nod,Z_nod,G_seg(1,:),x_refl,y_refl,z_refl,R); % apply Biot-Savart to calculate induced velocities. Function embedded at bottom of this script
    
    Ux = Ux_refl(1:length(x)) + Ux_refl(1+length(x):2*length(x)) + Ux_refl(1+2*length(x):3*length(x)) + Ux_refl(1+3*length(x):end); % use reflection and symmetry
    Uy = Uy_refl(1:length(x)) + Uy_refl(1+length(x):2*length(x)) - Uy_refl(1+2*length(x):3*length(x)) - Uy_refl(1+3*length(x):end); % use reflection and symmetry
    Uz = Uz_refl(1:length(x)) - Uz_refl(1+length(x):2*length(x)) + Uz_refl(1+2*length(x):3*length(x)) - Uz_refl(1+3*length(x):end); % use reflection and symmetry
    
    % plot instantaneous filament geometry and velocity vectors
    figure(f1)
    clf
    plot3(x_refl/d,z_refl/d,y_refl/d,'b.');
    hold on
    plot3(x_refl/d,z_refl/d,y_refl/d,'k-');
    axis equal
    axis([-1 6 -.5 .5 0 2])
    set(gca,'ydir','reverse')
    grid on
    box on
    quiver3(x/d,z/d,y/d,Ux/2,Uz/2,Uy/2,0,'color',[.6 .1 .1]);
    drawnow
    
    %% Update location of each node
    Uadv = repmat(adv_fact*Uinf,size(y_refl)); % constant advection velocity for whole filament
    
    dt = (tT(2) - tT(1))/(St*Uinf/d); % time-step
    
    X_nod = X_nod + (Ux + Uadv(1:length(x)))*dt; % update x-location using induced velocity and advection velocity
    Y_nod = Y_nod + Uy*dt; % update y-location using induced velocity
    Z_nod = Z_nod + Uz*dt; % update z-location using induced velocity
    
end

%% now calculate the velocity field
% describe data points - points at which induced velocity will be calculated
% Note that using too many points, in combination with too many filament
% elements can quickly increase RAM usage
x = 0*d:0.1*d:6*d;          % x-location of grid points
y = 0.1*d:0.1*d:2*d;      	% y-location of grid points
z = -0.1*d:0.1*d:0.1*d;     % z-location of grid points

[X,Y,Z] = meshgrid(x,y,z);

x_refl = [X(:)' X(:)'];     % reflection in the wall
y_refl = [Y(:)' Y(:)'];     % reflection in the wall
z_refl = [Z(:)' -Z(:)'];    % reflection in the wall

%% apply Biot-Savart
[Ux_refl,Uy_refl,Uz_refl] = Biot_Savart(X_nod,Y_nod,Z_nod,G_seg(1,:),x_refl,y_refl,z_refl,R);   % apply Biot-Savart to calculate induced velocities. Function embedded at bottom of this script

Ux = Ux_refl(1:length(Ux_refl)/2) + Ux_refl(1+length(Ux_refl)/2:end);   % reflection in the wall
Uy = Uy_refl(1:length(Ux_refl)/2) + Uy_refl(1+length(Ux_refl)/2:end);   % reflection in the wall
Uz = Uz_refl(1:length(Ux_refl)/2) - Uz_refl(1+length(Ux_refl)/2:end);   % reflection in the wall

Ux = reshape(Ux,[length(y),length(x),length(z)]);
Uy = reshape(Uy,[length(y),length(x),length(z)]);
Uz = reshape(Uz,[length(y),length(x),length(z)]);

%% plot velocity field
cm = hot(11);
cm = [fliplr(cm); flipud(cm)];

figure
subplot(211)
s1 = slice(X/d,Y/d,Z/d,Ux/10,[],[],[0]);
set(s1,'EdgeColor','none')
caxis([-.1 .1])
yy = get(s1, 'YData');
zz = get(s1, 'Zdata');
set(s1, 'YData', zz, 'ZData', yy);
axis equal
colorbar
hold on
plot3(X_nod/d,-Z_nod/d,Y_nod/d,'.k')

subplot(212)
s1 = slice(X/d,Y/d,Z/d,Uy/10,[],[],[0]);
set(s1,'EdgeColor','none')
caxis([-.1 .1])
yy = get(s1, 'YData');
zz = get(s1, 'Zdata');
set(s1, 'YData', zz, 'ZData', yy);
axis equal
colorbar
colormap(cm)
hold on
plot3(X_nod/d,-Z_nod/d,Y_nod/d,'.k')

%% function to build filaments
function [X_nod,Y_nod,Z_nod,G_seg] = Filament_Geometry(a,D,x0,y0,G,n)
%Creates a full filament description from vorticity data
% a: inclination angle (degrees)
% D: vortex-ring diameter
% x0,y0: starting position of vortex-ring center
% G:   Column vector containing circulation of vortices
% n:   Number of segments in the vertical head
%
% Creates a vortex ring when G(1)==G(2)
% Creates a vortex ring with legs when G(1)~=G(2)
% Creates a hairpin when G(1)==0

dx = D*cos(a/180*pi);
dy = D*sin(a/180*pi);
X = [x0-dx/2; x0+dx/2]';
Y = [y0-dy/2; y0+dy/2]';

M = [X Y G];
for i = 1:size(M,2)
    M(isnan(M(:,i)),:) = [];
end

D = sqrt((M(:,1)-M(:,2)).^2+(M(:,3)-M(:,4)).^2);
a = atan((M(:,4)-M(:,3))./(M(:,2)-M(:,1)));

C = [M(:,1)+(M(:,2)-M(:,1))/2 M(:,3)+(M(:,4)-M(:,3))/2];
X_sec = [M(:,1) M(:,2)];
Y_sec = [M(:,3) M(:,4)];
Z_sec = [zeros(size(M,1)) zeros(size(M,1))];
G_sec = [M(:,6) -M(:,5)];

l = pi*D/n;
X_leg = [C(:,1) C(:,1)-C(:,2)./tan(a)];
Y_leg = [C(:,2) zeros(size(C(:,2)))];
Z_leg = [D/2 D/2];
L_leg = sqrt((X_leg(:,2)-X_leg(:,1)).^2+(Y_leg(:,2)-Y_leg(:,1)).^2+(Z_leg(:,2)-Z_leg(:,1)).^2);
N_leg = round(L_leg./l);

da = pi/round(n/4)/2;
a_nod_1 = 0:da:pi/2;
a_nod_2 = pi/2:da:pi;
X_nod = single([]);
Y_nod = single([]);
Z_nod = single([]);
G_seg = single([]);

for F = 1:size(G_sec,1)
    X_nod_leg = linspace(X_leg(F,1),X_leg(F,2),N_leg(F));
    Y_nod_leg = linspace(Y_leg(F,1),Y_leg(F,2),N_leg(F));
    Z_nod_leg = linspace(Z_leg(F,1),Z_leg(F,2),N_leg(F));
%     G_seg_leg = repmat(G_sec(F,1)-G_sec(F,2),1,N_leg(F)); % legs with uniform circulation
    G_seg_leg = linspace(G_sec(F,1)-G_sec(F,2),0,N_leg(F)); % legs with decaying circulation
    
    X_nod = [X_nod NaN D(F)/2*cos(a_nod_1)*cos(a(F))+C(F,1) NaN D(F)/2*cos(a_nod_2)*cos(a(F))+C(F,1) NaN X_nod_leg(1:end-1)];
    Y_nod = [Y_nod NaN D(F)/2*cos(a_nod_1)*sin(a(F))+C(F,2) NaN D(F)/2*cos(a_nod_2)*sin(a(F))+C(F,2) NaN Y_nod_leg(1:end-1)];
    Z_nod = [Z_nod NaN D(F)/2*sin(a_nod_1) NaN D(F)/2*sin(a_nod_2) NaN Z_nod_leg(1:end-1)];
    G_seg = [G_seg [NaN repmat(G_sec(F,1),1,size(a_nod_1,2)) NaN repmat(G_sec(F,2),1,size(a_nod_2,2)) NaN G_seg_leg(1:end-1); NaN ones(1,size(a_nod_1,2)) NaN -ones(1,size(a_nod_2,2)) NaN zeros(size(G_seg_leg(1:end-1)))]];
end
end

%% function to calculate Biot-Savart
function [Ux,Uy,Uz] = Biot_Savart(X_nod,Y_nod,Z_nod,G_seg,x,y,z,R)
%Biot_Savart(X_nod,Y_nod,Z_nod,G_seg,X,Y,Z,R)
%Calculates the velocity induced by a filament described by X_nod, Y_nod,
%Z_nod and G_seg on point x,y,z. R represents Rankine vortex diameter

%%
X = x(:);
Y = y(:);
Z = z(:);

N_points = length(X);
N_seg = length(G_seg)-1;

x_seg = single(X_nod(2:end)/2+X_nod(1:end-1)/2); % midpoints of segments
y_seg = single(Y_nod(2:end)/2+Y_nod(1:end-1)/2);
z_seg = single(Z_nod(2:end)/2+Z_nod(1:end-1)/2);

r_vec(:,:,1) = (repmat(single(X),1,N_seg)-repmat(x_seg,N_points,1));  % distance (segment-datapoint) vector
r_vec(:,:,2) = (repmat(single(Y),1,N_seg)-repmat(y_seg,N_points,1));
r_vec(:,:,3) = (repmat(single(Z),1,N_seg)-repmat(z_seg,N_points,1));

clearvars X Y Z Xyz xYz xyZ x_seg y_seg z_seg

dl_vec = [(X_nod(2:end)-X_nod(1:end-1)); (Y_nod(2:end)-Y_nod(1:end-1)); (Z_nod(2:end)-Z_nod(1:end-1));];   % segment vector
dl_vec = permute(single(dl_vec),[3 2 1]);
n_vec(:,:,1) = repmat(dl_vec(:,:,2),N_points,1).*r_vec(:,:,3)-r_vec(:,:,2).*repmat(dl_vec(:,:,3),N_points,1); % vector perpendicular to r and dl
n_vec(:,:,2) = repmat(dl_vec(:,:,3),N_points,1).*r_vec(:,:,1)-r_vec(:,:,3).*repmat(dl_vec(:,:,1),N_points,1);
n_vec(:,:,3) = repmat(dl_vec(:,:,1),N_points,1).*r_vec(:,:,2)-r_vec(:,:,1).*repmat(dl_vec(:,:,2),N_points,1);
n_length = sqrt(n_vec(:,:,1).^2+n_vec(:,:,2).^2+n_vec(:,:,3).^2); % length of vector
n_unit(:,:,1) = n_vec(:,:,1)./n_length;  % unit vector perpendicular to dl and r
n_unit(:,:,2) = n_vec(:,:,2)./n_length;
n_unit(:,:,3) = n_vec(:,:,3)./n_length;
clearvars n_vec n_length

dl = sqrt(repmat(dl_vec(:,:,1),N_points,1).^2+repmat(dl_vec(:,:,2),N_points,1).^2+repmat(dl_vec(:,:,3),N_points,1).^2); % segment length
r = sqrt(r_vec(:,:,1).^2+r_vec(:,:,2).^2+r_vec(:,:,3).^2);  % distance length

costheta = (repmat(dl_vec(:,:,1),N_points,1).*r_vec(:,:,1)+repmat(dl_vec(:,:,2),N_points,1).*r_vec(:,:,2)+repmat(dl_vec(:,:,3),N_points,1).*r_vec(:,:,3))./(dl.*r);
clearvars dl_vec r_vec
costheta(costheta>1) = 1;
costheta(costheta<-1) = -1;
theta = acos(costheta); % angle (rad) between dl and r
clearvars costheta

G_seg = repmat(G_seg(1:end-1),N_points,1);
U(:,:,1) = G_seg/(4*pi) .* sin(theta)./(max(R,r).^2) .* dl .* n_unit(:,:,1); % calculate induced velocity
U(:,:,2) = G_seg/(4*pi) .* sin(theta)./(max(R,r).^2) .* dl .* n_unit(:,:,2);
U(:,:,3) = G_seg/(4*pi) .* sin(theta)./(max(R,r).^2) .* dl .* n_unit(:,:,3);

clearvars dl r n_unit theta

u_x = sum(U(:,:,1),2,'omitnan');
u_y = sum(U(:,:,2),2,'omitnan');
u_z = sum(U(:,:,3),2,'omitnan');

clearvars U
%%
Ux = reshape(u_x,size(x));
Uy = reshape(u_y,size(y));
Uz = reshape(u_z,size(z));

end